// $Id: ataxx.cc,v 1.40 2002/12/13 10:46:38 quarl Exp $

// ATAXX extensions (C) 2002 Karl Chen <quarl@hkn.eecs.berkeley.edu>.

//////////////////////////////////////////////////////////////////////
// ATAXX: BACKGROUND
//
// Rules for ATAXX can be found at:
//
//   http://braindamage.org/ataxx/rules.html
//   http://www.gamerz.net/pbmserv/ataxx.html#rules
//
// History:
//   http://www.pressibus.org/ataxx/gen/gborigines.html
//
//
//
//   Traditional board is 7x7 with each corner occupied:
//
//
//                 A B C D E F G
//               1 o . . . . . x
//               2 . . . . . . .
//               3 . . . . . . .
//               4 . . . . . . .
//               5 . . . . . . .
//               6 . . . . . . .
//               7 x . . . . . o
//
//
//
//
// A move consists of either a "growth" or a "jump". A growth is when you
// select one of your pieces and place a new piece in an empty adjacent space,
// including diagonally. A jump is when you select one of your pieces and move
// it to an empty space that is two position away. This includes diagonals and
// "knights moves".
//
//   j j j j j
//   j g g g j
//   j g . g j    g = valid "growth" destinations from the center
//   j g g g j    j = valid "jump" destinations form the center
//   j j j j j

// Once placed, any opponents pieces that are adjacent to the new piece
// (either through growth or jump) are flipped the to the moving player's
// colour.
//
//
// NON-NORMAL-FORM: continued moves
//
// If you can't move, your turn is simply skipped. This is different from
// normal form; we implement it by defining an ATAXX MOVE as a sequence of
// plies until your opponent can move.

//////////////////////////////////////////////////////////////////////
// ALGORITHM & ANALYSIS
//
//
//    LOOPS ARE IGNORED
//
//    ATAXX is a potentially loopy game. Loops can happen if both players jump
//    back and forth ad infinitum. However, they can only happen if both
//    players desire to continue looping. We will assume that the player that
//    has the better position (without the potential loop) will avoid it.
//
//
//    TINIES?
//
//    I haven't discovered any tinies yet. Most games are switches or
//    switches of switches ad nauseum.
//
//    You can get a star like this:j
//
//           +-----+
//           |X   O| *
//           +-----+
//
//       or
//
//
//
//    DECOMPOSITION
//
//    Splitting games into multiple components is difficult except in rare
//    and contrived cases.
//
//
//    SOLVING
//
//    Attempting to solve the starting 7x7 board is futile. I suspect that it
//    is some kind of switch with mean of 0, and temperature up to 49, and
//    that any sufficiently large board is like this.
//
//    A 7x7 board has about 3^49 = 2.4x10^23 positions. Huge!  Even a 5x5
//    board is too large. 3^25 = 8.5x10^11 positions. Each board takes 50
//    bits; rounding up to the nearest word, 8 bytes. And that's just the
//    position hashing, nevermind the game hashing! Too large for today's
//    (2002) computing powers. Maybe in a couple years it'll be child's play
//    though?
//
//    4x4 is easily done at about 200 MB total RAM used.
//

//////////////////////////////////////////////////////////////////////
// CODE COMMENTARY
//
//    ENTRY POINTS from 'games':
//
//       game_type ataxx();
//
//          Inputs an ATAXX position from stdin and returns its value.
//
//       void test_ataxx(game_type);
//
//          Inputs an ATAXX position from stdin and compares to $1. Useful for
//          test scripts.
//
//          the "ATAXXDEBUG" game variable controls whether game values are
//          printed by test_ataxx() as they are calculated.
//
//
//    BITBOARDS
//
//    Instead of using bitmatrx.[ch], I wrote new bit-board routines because
//    1) the C macros suck 2) we don't need the extra board size tracking in
//    this module 3) hashing of wolfe's bitboards is done by converting a
//    bitboard to a fucking linked list 4) it's not very hard to do it better
//    5) we need actually two bits of information, and it's better to have a
//    single data structure with single access points, than two bitboards.
//
//    point 3 brings us to...
//
//    HASHING
//
//    Obviously, positions need to be hashed and cached. Wolfe's hashing sucks
//    though because it uses a single hash table for the entire program, and
//    thus needs to convert every key into a fucking linked list. and then all
//    those extra sub-program keys to differentiate between game modules --
//    ugghh, inefficient.
//
//    We'll rely here on g++'s hash table. The hashcode algorithm is the same
//    as Wolfe's except that we don't convert first to a list, and thus is a
//    zillion times faster.
//

// BLUE = BLACK = LEFT  = X
// RED  = WHITE = RIGHT = O
// GREEN       = BLOCKS = #
//               EMPTY  =   (space)

extern "C" {
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "gameops.h"
#include "output.h"
}

// plus() and minus() from "gameops.h" and <algorithm> clash.  namespace
// support isn't great yet - just #define them away since we don't need them.
#define plus std_plus
#define minus std_minus
#include <cstdio>
#include <iostream>
#include <string>
#include <list>
#include <hash_map>
#undef plus
#undef minus

using namespace std;

extern "C" {
    game_type ataxx();
    void test_ataxx(game_type);
}

#define hashA ((unsigned long) 67343)
#define hashB ((unsigned long) 72763)

// implements a "double-bit" board.
class DBitBoard
{
public:
    DBitBoard() : board(NULL) {}
    // an empty board is allowed initially to support AtaxxPosition.Input()
    // not getting parameters until later. we shouldn't have empty boards
    // anywhere else, though.

    DBitBoard(DBitBoard const& b) : board(b.board) { assert (board); ++refCount(); }
    ~DBitBoard() { destroy(); }

    DBitBoard& operator= (DBitBoard const& b)
    {
        if (&b != this) {
            destroy();
            board = b.board;
            ++refCount();
        }
        return *this;
    }

    DBitBoard clone() const
    {
        DBitBoard b;
        b.internal_allocate(getBBSize(), xSizeParams());
        memcpy(b.xNonVolatileData(), xNonVolatileData(), xNVTSize());
        return b;
    }

    bool operator ==(DBitBoard const& b) const
    {
        // TODO: allow comparing two empty boards?
        assert (board);
        assert (b.board);

        return
            board == b.board ||
            0==memcmp(b.xNonVolatileData(), xNonVolatileData(), xNVTSize());
    }

    unsigned long hash() const
    {
        assert (board != NULL);

        unsigned long val = 0;

        for (bbtype const* p = xNonVolatileData();
             p != board+xNVTSize()/sizeof(bbtype); ++p)
        {
            val = hashA * (val + *p + 1) + hashB;
        }

        return val;
    }

    void allocate(int nRows, int nCols)
    {
        assert (board == NULL);

        assert (nRows > 0 && nCols > 0);
        assert (nRows < 256 && nCols < 256);

        int nBits = nCols * nRows * 2;
        int nBytes = (nBits+7) / 8;
        int bbSize = (nBytes + sizeof(bbtype) - 1) / sizeof(bbtype);

        internal_allocate(bbSize, bbSize << 16 | (nRows << 8) | nCols);
    }

    DBitBoard Rotate() const;
    DBitBoard InvertBit2() const;

public:
    int getBBSize() const { return xSizeParams() >> 16; }
    int getRows() const { return (xSizeParams() >> 8) & 255; }
    int getCols() const { return xSizeParams() & 255; }

    //static int getRows() { return nRows; }
    //static int getCols() { return nCols; }

    //static void setDimensions(int rows, int cols)
    //{
    //    nRows = rows;
    //    nCols = cols;

    //    int nBits = nCols * nRows * 2;
    //    int nBytes = (nBits+7) / 8;
    //    bbSize = (nBytes + sizeof(bbtype) - 1) / sizeof(bbtype);
    //}

public:
    bool inRange(int row, int col) const
    {
        return
            row >= 0 && row < getRows() &&
            col >= 0 && col < getCols();
    }

    int getDBit(int row, int col) const
    {
        assert (inRange(row, col));
        assert (board);

        return ( byte(row, col) & bitmask(row, col) ) >> bit(row, col);
    }

    void setDBit(int row, int col, int dbit)
    {
        assert (inRange(row, col));
        assert (board);

        // must contain only the two lowest bits
        assert (dbit == dbit & 3);

        byte(row, col) = byte(row, col) & ~(bitmask(row, col)) | dbit << bit(row, col);
    }

protected:
    typedef unsigned bbtype;

    int spot(int row, int col) const
    {
        return 2 * (row * getCols() + col);
    }

    bbtype& byte(int row, int col)
    {
        return board[spot(row,col) / (sizeof(bbtype)*8)];
    }

    bbtype byte(int row, int col) const
    {
        return board[spot(row,col) / (sizeof(bbtype)*8)];
    }

    int bit(int row, int col) const
    {
        return spot(row,col) % (sizeof(bbtype)*8);
    }

    int bitmask(int row, int col) const
    {
        return 3 << bit(row, col);
    }

    void internal_allocate(unsigned bbSize, unsigned sizeParam)
    {
        assert (board == NULL);

        board = new bbtype[bbSize+2] + 2;
        board[-1] = sizeParam;
        refCount() = 1;
    }

    void destroy()
    {
        if (board) {
            if (--refCount() == 0) {
                delete [] (board-2);
            }
        }
    }

    bbtype& refCount() const { assert(board); return *const_cast<bbtype*>(board-2); }

    unsigned xSizeParams() const { assert(board); return board[-1]; }

    bbtype* xNonVolatileData() { return board-1; }
    bbtype const* xNonVolatileData() const { return board-1; }
    size_t xNVTSize() const { return (getBBSize()+1) * sizeof(bbtype); }


protected:
    bbtype* board;

    //static int nRows, nCols, bbSize;
};

DBitBoard DBitBoard::Rotate() const
{
    DBitBoard b;
    b.internal_allocate(getBBSize(), xSizeParams());

    for (int r = 0; r < getRows(); ++r)
    {
        for (int c = 0; c < getCols(); ++c)
        {
            b.setDBit( getRows()-r-1, getCols()-c-1, getDBit(r, c) );
        }
    }

    return b;
}

inline int InvertBit2(int n)
{
    switch (n) {
    case 0: return 0;
    case 1: return 1;
    case 2: return 3;
    case 3: return 2;
    default: assert(0); return n;
    }
}

DBitBoard DBitBoard::InvertBit2() const
{
    DBitBoard b;
    b.internal_allocate(getBBSize(), xSizeParams());

    for (int r = 0; r < getRows(); ++r)
    {
        for (int c = 0; c < getCols(); ++c)
        {
            b.setDBit( r, c, ::InvertBit2(getDBit(r, c)));
        }
    }

    return b;
}

// this is pretty dumb and totally not re-entrant. but we only have one game
// at a time and this makes it so much easier! the rest of Gamesman's Toolkit
// is all nonreentrantlyC anyway.
//int DBitBoard::nRows = 0;
//int DBitBoard::nCols = 0;
//int DBitBoard::bbSize = 0;

class OptionList;

class AtaxxPosition
{
public:
    AtaxxPosition() {}
    AtaxxPosition(DBitBoard const& b) : position(b) {}
    AtaxxPosition(AtaxxPosition const& p) : position(p.position) {}
    ~AtaxxPosition() {}

    // input a position from stdin. returns true on successful input.
    bool Input();

    // return position score.
    game_type Score();

    // print to stdout the best move(s) for left and right
    void ShowBestMoves() const;

    void Print(const char* prefix = "", const char* line2 = 0) const;
    void Print(const char* prefix, string const& s) const { Print(prefix, s.c_str()); }

protected:

    enum pieceTYPES {
        pieceEMPTY = 0, pieceBLOCK = 1, pieceO = 2, pieceX = 3
    };

    static char pieceToChar(int p)
    {
        switch (p) {
        case pieceEMPTY: return ' ';
        case pieceBLOCK: return '#';
        case pieceO: return 'O';
        case pieceX: return 'X';
        default: assert(0 && "internal error"); return '?';
        }
    }

    static int charToPiece(char c)
    {
        switch (c) {
        case ' ': return 0;
        case '#': return 1;
        case 'o':
        case 'O': return 2;
        case 'x':
        case 'X': return 3;
        default: assert (0 && "internal error"); return 0;
        }
    }

    game_type ScoreNoHashing() const;
    int PlayAllOptions(OptionList& lOptions, OptionList& rOptions) const;
    void PlayMovesTo(OptionList& lOptions, OptionList& rOptions, int rTo, int cTo) const;
    void TryMoveFromTo(bool& lTried, bool& rTried,
                       OptionList& lOptions, OptionList& rOptions,
                       int rFrom, int cFrom,
                       int rTo, int cTo,
                       bool bJump) const;
    int MakeMove(int rTo, int cTo, int rFrom, int cFrom, bool bJump);

protected:

    DBitBoard position;

    struct DBitBoard_Hash {
        unsigned long operator()(DBitBoard b) const {
            return b.hash();
        }
    };

    // use the g++ (SGI/HP) version of hash_map
    typedef hash_map<DBitBoard, game_type, DBitBoard_Hash> PosHashtable;

    static PosHashtable hashtable;

    enum TransformType { ttNone, ttRotate, ttRotateInvert };

    static TransformType transform;

};

AtaxxPosition::PosHashtable AtaxxPosition::hashtable;
AtaxxPosition::TransformType AtaxxPosition::transform = AtaxxPosition::ttNone;
bool debugPrint = 0;

// entry point from Gamesman's Toolkit.
game_type ataxx()
{
    AtaxxPosition pos;
    game_type r = pos.Input() ? pos.Score() : zero;
    assert (r);
    return r;
}

inline string gameToString(game_type g)
{
    char buf[MAXBUFF];
    return g ? game_sprintn(buf, g, sizeof(buf)-1) : "NO_GAME";
}

int nTest = 0;
int nPass = 0;

// non-error, quiet version of get_var
inline game_type get_var_or_null (var_type v)
{
    if (hash_test (VAR_KEY, v))
        return hash_get_last();
    else {
        return NO_GAME;
    }
}

inline bool var_exists_and_nonzero(var_type v)
{
    game_type g = get_var_or_null(v);

    return g && !eq(g, zero);
}

static var_type debugvar = make_var("ATAXXDEBUG");

// entry point from Gamesman's Toolkit.
void test_ataxx(game_type g)
{
    ++nTest;

    AtaxxPosition pos;
    if (!pos.Input()) {
        return;
    }

    //pos.Print();

    game_type h = pos.Score();

    if (g == h) {
        cout << "## TEST " << nTest << " PASSED: " << gameToString(g) << endl;
        ++nPass;
    } else {
        cout << "## TEST " << nTest << " FAILED: expected "
             << gameToString(g) << ", calculated " << gameToString(h) << endl;
    }
    cout << "----------------------------------------------------------------------\n";

}


inline bool isVerticalEdge(char c)
{
    return c == '+' || c == '|';
}

inline bool containsOnlyHorizontalEdges(string const& s)
{
    assert (!s.empty());
    return
        s.find_first_not_of("-+") == string::npos;
}

// make sure row contains only valid chars, and lowercase them as neccessary.
inline bool validateRow(string const& s)
{
    return s.find_first_not_of(" #XxOo") == string::npos;
}

bool AtaxxPosition::Input()
{
    if (flags & ECHO_INPUT_FLAG) {
        cout << '\n';
    }
    if (flags & PROMPT_FLAG) {
        cout << "Enter ATAXX position (<sp>, #, X, O; borders +) followed by extra <cr>\n";
    }

    debugPrint = var_exists_and_nonzero(debugvar);

    // these used to be state variables but they're not needed in score
    // calculation - there are always edges in ATAXX. these variables are now
    // used only for input parsing.
    bool northEdge = false, southEdge = false, eastEdge = false, westEdge = false;

    int nCols = 0;

    typedef list<string> StringList;

    StringList rows;
    string line;
    while (getline(cin, line) && !line.empty()) {
        rows.push_back(line);
        nCols = max(nCols, (int) line.size());
    }

    int nRows = rows.size();

    if (rows.empty()) return false;

    // check north edge
    if (containsOnlyHorizontalEdges(rows.front())) {
        northEdge = true;
        rows.pop_front();
    }

    if (rows.empty()) return false;

    // check south edge
    if (containsOnlyHorizontalEdges(rows.back())) {
        southEdge = true;
        rows.pop_back();
    }

    if (rows.empty()) return false;

    // check east/west edges
    {
        string& firstRow = rows.front();

        westEdge = isVerticalEdge(firstRow[0]);
        eastEdge = ((int) firstRow.size() == nCols) &&
            isVerticalEdge(firstRow[nCols-1]);

        if (nCols - eastEdge - westEdge <= 0) return false;

        int n = 0;

        for (StringList::iterator i = rows.begin();
             i != rows.end();
             ++i)
        {
            ++n;
            string& row = *i;

            if (eastEdge) {
                // make sure east edge exists.
                if ((int) row.size() < nCols ||
                    !isVerticalEdge(row[nCols-1]))
                {
                    cerr << "Invalid character or not enough characters in row "
                         << n << " column " << row.size() << " (missing right edge)\n";
                    return false; // FAILED!
                }
                row.erase(nCols-1,1);
            } else {
                // pad west if neccessary, since no edge
                row.append(nCols-row.size(), ' ');
            }

            if (westEdge) {
                // make sure west edge exists.
                if (!isVerticalEdge(row[0])) {
                    cerr << "Invalid character in row " << n << " column 0 (missing left edge)\n";
                    return false; // FAILED!
                }
                row.erase(0,1);
            }

            if (!validateRow(row)) {
                cerr << "Invalid character(s) in row " << n << " '" << row << "'\n";
                return false;   // FAILED!
            }
        }

    }

    nRows -= (northEdge + southEdge);
    nCols -= (eastEdge + westEdge);

    //DBitBoard::setDimensions(nRows, nCols);

    //printf("nesw = %d%d%d%d; [r,c]=[%d,%d]\n",
    //       northEdge,eastEdge,southEdge,westEdge,
    //       nRows, nCols);

    // initialize the DBitBoard
    position.allocate(nRows, nCols);
    {
        int r = 0;
        for (StringList::iterator i = rows.begin();
             i != rows.end();
             ++i, ++r)
        {
            string& row = *i;

            for (int c = 0; c < (int) row.size(); ++c) {
                position.setDBit(r, c, charToPiece(row[c]));
            }
        }
    }

    // check to see if this board is equal to its rotation or its inverted
    // rotation. if it is we'll check each position against its
    // rotation/inversion when hashing.
    DBitBoard boardRotated = position.Rotate();
    if (boardRotated == position) {
        transform = ttRotate;
    } else {
        DBitBoard boardRotatedInverted = boardRotated.InvertBit2();
        if (boardRotatedInverted == position) {
            transform = ttRotateInvert;
        } else {
            transform = ttNone;
        }
    }

    if (flags & ECHO_INPUT_FLAG) {
        Print();
    }

    //if (debugPrint) {
    //    cout << "## transform = " << transform << endl;
    //}

    return true;
}

void AtaxxPosition::Print(const char* prefix, const char* line2) const
{
    cout << prefix << '+' << string(position.getCols(), '-') << '+' << '\n';

    for (int r = 0; r < position.getRows(); ++r)
    {
        cout << prefix << '|';
        for (int c = 0; c < position.getCols(); ++c)
        {
            cout << pieceToChar(position.getDBit(r, c));
        }
        cout << '|';
        if (r == 0 && line2) cout << ' ' << line2;
        cout << '\n';
    }

    cout << prefix << '+' << string(position.getCols(), '-') << '+' << '\n';

    //cout << "hash = " << hash() << endl;
}

//inline string itostring(int x)
//{
//    char buf[33];
//    sprintf(buf, "%d", x);
//    return buf;
//}

#define LOOPY_GAME NO_GAME

game_type AtaxxPosition::Score()
{
    //game_type& hashed_value = hashtable[*this];

    //if (hashed_value == NO_GAME) {
    //    hashed_value = ScoreNoHashing();
    //}

    //Print("              ", "[calculating]");

    pair<PosHashtable::iterator, bool> p =
        hashtable.insert( PosHashtable::value_type( position, LOOPY_GAME ) );

    game_type& hashed_value = p.first->second;

    if (p.second) {
        // inserted new value. that means we've never seen this game before.

        // optimization: check for the position rotated.
        //
        // we don't keep the rotate position in the hash because we actually
        // need to save memory a lot more than cpu.

        if (transform == ttRotate) {

            PosHashtable::const_iterator i = hashtable.find(position.Rotate());
            if (i != hashtable.end() && i != p.first) {
                // found rotation.
                if (debugPrint)
                    Print("   ", "(hashed: rotated) " + gameToString(i->second));

                // delete the one we just added since we don't want it
                // anymore. If we had lots of RAM and wanted to optimize CPU, we
                // could keep it and assign the same value to it.

                hashtable.erase(p.first);
                return i->second;
            }
        } else if (transform == ttRotateInvert) {

            //Print("        ORIGINAL: ", "hash="+itostring(hash()));
            //AtaxxPosition DEBUG_IR = (*(AtaxxPosition*)(&Rotate().InvertBit2()));
            //DEBUG_IR.Print("INVERTED ROTATED: ", "hash="+itostring(DEBUG_IR.hash()));

            PosHashtable::const_iterator i = hashtable.find(position.Rotate().InvertBit2());
            if (i != hashtable.end() && i != p.first) {

                if (debugPrint)
                    Print("   ", "(hashed: inverted rotated) " + gameToString(negative(i->second)));

                hashtable.erase(p.first);
                return negative(i->second);
            }

        }


        hashed_value = ScoreNoHashing();

        if (debugPrint) {
            Print("   ", gameToString(hashed_value).c_str());
        }

    } else {
        // no value inserted; it already existed. there are two cases for the
        // hashed value:
        //
        //   1) LOOPY_GAME : we already inserted this game into the hashtable
        //                   but haven't set its value yet. That means it's
        //                   LOOPY! Stop processing it. we just return
        //                   LOOPY_GAME to indicate this, so we're done.
        //
        //   2) else       : we calculated its score already and we're done.

        if (debugPrint)
            Print("   ", "(hashed) " + gameToString(hashed_value));

        // assign to this one (if it's not already the same one) so that we
        // keep only copy of the data.
        position = p.first->first;
    }

    return hashed_value;
}

//struct BoardAndValue
//{
//    AtaxxPosition pos;
//    game_type value;
//};

// we use class inheritance rather than simple typedef here because it makes
// it easier to pre-declare it above.

// we use a hash to store the score but not because we want to remember the
// score (there's already a hashtable for that), but just so that we only have
// one copy of the position in the OptionsList (we used to use a list<> which
// would have repeats). a hashset<> or even set<> would be have worked too but
// this was easier to write.

class OptionList : public list<pair<AtaxxPosition,game_type> > //BBHashtable
{
public:

    // reverses the list in the process, but doesn't matter for our purposes.
    list_type to_value_list() const
    {
        list_type gl = list_make();

        for (const_iterator i = begin();
             i != end();
             ++i)
        {
            //game_type g = i->value;
            game_type g = i->second;
            list_prepend(gl, g);
        }

        return gl;
    }


    // insert an option unless it is <= an option that already exists.
    void MaybeInsert(AtaxxPosition const& pos, game_type score)
    {
        // does something at least as good exist already?

        if (1) for (const_iterator i = begin();
             i != end();
             ++i)
        {
            if (ge(i->second, score))
                return;
        }

        // are we better than something else?
        for (iterator i = begin();
             i != end();
             )
        {
            if (gt(score, i->second)) {
                iterator j = i++;
                erase(j);
                continue;
            }

            ++i;
        }

        // put it in now.
        Insert(pos, score);

        //(*this) [newPos] = score;
    }

    void Insert(AtaxxPosition const& pos, game_type score)
    {
        push_back(make_pair(pos, score));
    }


};

// this is called by Score() after hashing has already been taken into
// account.
game_type AtaxxPosition::ScoreNoHashing() const
{
    OptionList lOptions, rOptions;
    OptionList possibleMoves;

    possibleMoves.Insert(*this, LOOPY_GAME);

    int lCount = 0;
    int rCount = 0;
    int sCount = 0;
    for (int r = 0; r < position.getRows(); ++r) {
        for (int c = 0; c < position.getCols(); ++c) {
            switch (position.getDBit(r, c)) {
            case pieceX: ++lCount; break;
            case pieceO: ++rCount; break;
            case pieceEMPTY: ++sCount; break;
            }
        }
    }

    if (sCount == 0) {
        // no moves left. return signed integer value.
        return g_int(lCount - rCount);
    }

    if (lCount == 0) {
        // left has no pieces! right owns the board!
        return g_int(-(rCount + sCount));
    }

    if (rCount == 0) {
        // ditto for vice versa
        return g_int(lCount + sCount);
    }

    do {
        //cerr << "## " << possibleMoves.size() << " possible moves\n";

        assert (!possibleMoves.empty());

        int signedCount = 0;

        for (OptionList::const_iterator i = possibleMoves.begin();
             i != possibleMoves.end();
             ++i)
        {
            signedCount = i->first.PlayAllOptions(lOptions, rOptions);
        }

        //cerr << "## " << lOptions.size() << " lOptions, " << rOptions.size() << " rOptions\n";

        if (lOptions.empty() && rOptions.empty()) {
            // this is the base case of our huge recursion cycle.
            return g_int(signedCount);
        }

        // if only one player can move, then the value of this position is the option
        // for the player that can move. this is because the rules of ATAXX
        // say that if one player can't move the other moves.
        //

        // we can't just set e.g. left's options to right's options if his
        // are empty, because he doesn't get to choose from them!

        game_type kludgeval = zero;

        if (lOptions.empty()) {
            kludgeval = one;
            //list_type l = rOptions.to_value_list();
            //return plus(make(l, zero), ); // KLUDGE
        }

        if (rOptions.empty()) {
            kludgeval = neg1;
            //list_type l = lOptions.to_value_list();
            //return plus(make(l, l), star(1)); // KLUDGE
        }

        // The way we implement this is by continuing to replace moves with
        // their children, until either we run out of moves (see above:
        // return an integer), or both players have options.

        //if (lOptions.empty()) {
        //    possibleMoves.clear();
        //    rOptions.swap(possibleMoves);
        //    continue;
        //}
        //if (rOptions.empty()) {
        //    possibleMoves.clear();
        //    lOptions.swap(possibleMoves);
        //    continue;
        //}

        return plus(make(lOptions.to_value_list(), rOptions.to_value_list()), kludgeval);
    } while (1);
}


inline int oppPiece(int piece)
{
    switch (piece) {
    case 2:
        return 3;
    case 3:
        return 2;
    default:
        assert (0); return 0;
    }
}

// returns piece that moved.
int AtaxxPosition::MakeMove(int rFrom, int cFrom, int rTo, int cTo, bool bJump)
{
    assert (position.inRange(rFrom, cFrom));
    assert (position.inRange(rTo, cTo));
    assert (position.getDBit(rTo, cTo) == pieceEMPTY);

    //printf("## MakeMove( [%d,%d] -> [%d,%d] )\n", rFrom, cFrom, rTo, cTo);

    int piece = position.getDBit(rFrom, cFrom);
    assert (piece == pieceX || piece == pieceO);

    assert (bJump ==
            (rTo-rFrom > 1 || rTo-rFrom < -1 ||
             cTo-cFrom > 1 || cTo-cFrom < -1 ) );

    if (bJump)
    {
        position.setDBit(rFrom, cFrom, pieceEMPTY);
    }

    // put new piece in
    position.setDBit(rTo, cTo, piece);

    // flip all opponent's pieces.
    int opp = oppPiece(piece);

    for (int r = rTo-1; r <= rTo+1; ++r) {
        for (int c = cTo-1; c <= cTo+1; ++c) {
            if (position.inRange(r, c) && position.getDBit(r, c) == opp)
                position.setDBit(r, c, piece);
        }
    }

    return piece;
}

int AtaxxPosition::PlayAllOptions(OptionList& lOptions, OptionList& rOptions) const
{
    int signedCount = 0;

    for (int rTo = 0; rTo < position.getRows(); ++rTo) {
        for (int cTo = 0; cTo < position.getCols(); ++cTo) {
            // do fancy shit.

            int piece = position.getDBit(rTo, cTo);

            switch (piece) {
            case pieceEMPTY:
                PlayMovesTo(lOptions, rOptions, rTo, cTo);
                break;

            case pieceX:
                signedCount++;
                break;

            case pieceO:
                signedCount--;
                break;

            case pieceBLOCK:
                break;

            default:
                assert(0 && "internal error");
            }
        }
    }

    return signedCount;
}

inline void AtaxxPosition::TryMoveFromTo(bool& lTried, bool& rTried,
                                         OptionList& lOptions, OptionList& rOptions,
                                         int rFrom, int cFrom,
                                         int rTo, int cTo,
                                         bool bJump) const
{
    if (rFrom == rTo && cFrom == cTo) return;
    if (!position.inRange(rFrom, cFrom)) return;

    int piece = position.getDBit(rFrom, cFrom);

    if (piece <= 1) return;
    if (piece == pieceX && lTried) return;
    if (piece == pieceO && rTried) return;

    AtaxxPosition newPos(position.clone());

    int pieceMoved = newPos.MakeMove(rFrom, cFrom, rTo, cTo, bJump);

    // calling Score() puts it into the hash table or replaces it with its
    // equivalent.

    game_type score = newPos.Score();
    //bav.value = bav.pos.Score();

    if (score != LOOPY_GAME) {
        switch (pieceMoved) {
        case pieceX:
            //lOptions.push_back( bav );
            // assert that if we've added this position before that
            // the score hasn't changed. should be a trivial
            // assertion since Score() is hashed too.
            //assert (lOptions.find(newPos)==lOptions.end() || lOptions[newPos] == score);

            lOptions.MaybeInsert(newPos, score);
            if (!bJump) lTried = true;
            break;

        case pieceO:
            //rOptions.push_back( bav );
            //assert (rOptions.find(newPos)==rOptions.end() || rOptions[newPos] == score);
            //rOptions[newPos] = score;

            rOptions.MaybeInsert(newPos, score);
            if (!bJump) rTried = true;
            break;
        }
    }
}

// play into the empty square (r,c). puts left's moves into lOptions, right's
// into rOptions.
void AtaxxPosition::PlayMovesTo(OptionList& lOptions, OptionList& rOptions,
                                int rTo, int cTo) const
{
    assert (position.inRange(rTo, cTo));
    assert (position.getDBit(rTo, cTo) == pieceEMPTY);

    // TODO: is it safe to assume that it is always better to extend to an
    // empty square than to jump to that square?

    bool lTried = false;
    bool rTried = false;

    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-1, cTo-1, rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo  , cTo-1, rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+1, cTo-1, rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-1, cTo  , rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+1, cTo  , rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-1, cTo+1, rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo  , cTo+1, rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+1, cTo+1, rTo, cTo, false);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-2, cTo-2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-2, cTo-1, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-2, cTo  , rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-2, cTo+1, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-2, cTo+2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+2, cTo-2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+2, cTo-1, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+2, cTo  , rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+2, cTo+1, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+2, cTo+2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-1, cTo-2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo  , cTo-2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+1, cTo-2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo-1, cTo+2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo  , cTo+2, rTo, cTo, true);
    TryMoveFromTo(lTried, rTried, lOptions, rOptions, rTo+1, cTo+2, rTo, cTo, true);

}

